package de.fraunhofer.iem.dataprovider.toolRun.service

import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.tool.dto.ToolDto
import de.fraunhofer.iem.dataprovider.tool.enumeration.ToolType
import de.fraunhofer.iem.dataprovider.toolRun.dto.ToolRunDto
import de.fraunhofer.iem.dataprovider.toolRun.entity.ToolRunEntity
import de.fraunhofer.iem.dataprovider.toolRun.repository.ToolRunRepository
import de.fraunhofer.iem.dataprovider.tools.gitlab.service.RepositoryDetailsService
import de.fraunhofer.iem.dataprovider.tools.occmd.service.OccmdService
import de.fraunhofer.iem.dataprovider.tools.ort.service.OrtService
import kotlinx.coroutines.*
import org.springframework.stereotype.Service

@Service
class ToolRunService(
    private val toolRunRepository: ToolRunRepository,
    private val repositoryService: RepositoryService,
    private val repositoryDetailsService: RepositoryDetailsService,
    private val kpiService: KPIService,
    private val ortService: OrtService,
    private val occmdService: OccmdService
) {

    private val defaultScope = CoroutineScope(Dispatchers.Default)
    private val ioScope = CoroutineScope(Dispatchers.IO)
    private val logger = getLogger(javaClass)

    /**
     * Warning: Long running function!
     *
     * In this function we calculate the KPIs for the given project using information from the gitlab api.
     * Therefore, we first create a repository and tool run object in the DB.
     *
     * Next, we asynchronously query the gitlab API and dedicated opencode tool APIs for further data.
     * Based upon this data we calculate RAW KPIs.
     *
     * Once all API responses have been processes we create a KPI tree and store this to the DB.
     *
     * Lastly, we update the tool run with all tools, which provided results for the given repository,
     * and store this to the DB.
     */
    suspend fun createToolRunForRepository(repoId: Long) = defaultScope.launch {
        /**
         * Create initial db models.
         * If this fails we want the whole method to fail as we can't continue
         * without the db models.
         *
         */
        logger.info("Starting tool run creation for repo $repoId")
        val toolRun = async(ioScope.coroutineContext) {
            val repoDto = repositoryService.getRepositoryInfo(repoId)
            val repo = repositoryService.getOrCreate(repoDto)
            val tr = createToolRunForRepository(repo)
            tr
        }.await()
        logger.info("Finished tool run creation for repo $repoId $toolRun")

        /**
         * For all tools, we in parallel query the tool results API.
         * If the tool has results, add the tool to the tool run
         * and then calculate the raw KPI based on the results.
         * When all tool APIs are processed we purge the old KPIs
         * from the db and store the new results.
         * Lastly, we update the tool run object with all tools,
         * which returned results for this repository.
         */
        logger.info("Starting batch API query for repo $repoId")
        val apiJobs = listOf(
            async {
                val vulnerabilityDtos =
                    ortService.getOrtResults(106) // in the dev setup we get results for repo id 106
                if (vulnerabilityDtos.isNotEmpty()) {
                    toolRun.toolEntities.add(ortService.toolEntity)
                }
                kpiService.calculateVulnerabilityKpis(vulnerabilityDtos)
            },

            async {
                val repoDetailsDto = repositoryDetailsService.getRepositoryDetails(repoId)
                toolRun.toolEntities.add(repositoryDetailsService.toolEntity)
                kpiService.calculateRepositoryDetailsKpis(repoDetailsDto)
            },

            async {
                val rawOccmdResults = occmdService.runOccmd(repoId, toolRun.repository.url)
                if (rawOccmdResults.isNotEmpty()) {
                    toolRun.toolEntities.add(occmdService.toolEntity)
                }
                kpiService.calculateOccmdKpis(rawOccmdResults)
            }
        )

        /**
         * Jobs created with async don't fail inside the async block, but at the place where await
         * is called. Thus, we warp the await call with a try catch and remove all failed results with
         * the surrounding mapNotNull. This process can be used for all optional API calls.
         */
        val kpis = apiJobs.mapNotNull {
            try {
                it.await()
            } catch (exception: Exception) {
                logger.error("API job failed with error $exception")
                null
            }
        }.flatten()

        logger.info("All API queries finished for repo $repoId. Calculated ${kpis.size} raw KPIs.")

        if (kpis.isNotEmpty()) {
            logger.info("Purging and storing new KPIs for repo $repoId")
            kpiService.storeAndPurgeOld(toolRun.repository, kpis)
            saveToolRunEntity(toolRun)
        }

        logger.info("Finished processing repository changed request for repository $repoId")
    }

    fun createToolRunForRepository(repo: RepositoryEntity): ToolRunEntity {
        val tr = ToolRunEntity(repository = repo)
        return toolRunRepository.save(tr)
    }

    fun saveToolRunEntity(toolRunEntity: ToolRunEntity): ToolRunEntity {
        return toolRunRepository.save(toolRunEntity)
    }

    suspend fun getToolRunForRepository(
        repo: RepositoryEntity,
        dispatcher: CoroutineDispatcher = Dispatchers.IO
    ): ToolRunDto {
        val toolRunEntity = withContext(dispatcher) {
            toolRunRepository.findByRepositoryId(repo.id!!)
        }

        val apiJobs: MutableList<Deferred<ToolDto>> = mutableListOf()

        toolRunEntity.toolEntities.forEach { tool ->
            when (tool.toolType) {
                ToolType.ORT -> {
                    apiJobs.add(
                        defaultScope.async {
                            // TODO: this will be replaced by an API call
                            val rawOrtResult = ortService.getOrtResults(106)
                            val findings = ortService.getFindings(rawOrtResult)
                            ToolDto(tool.name, findings)
                        }
                    )
                }

                else -> {
                    apiJobs.add(
                        defaultScope.async {
                            ToolDto(tool.name)
                        }
                    )
                }
            }
        }

        val tools = apiJobs.mapNotNull {
            try {
                it.await()
            } catch (exception: Exception) {
                logger.error("API job failed with error $exception")
                null
            }
        }

        return ToolRunDto.getDtoFromEntityWithToolFindings(toolRunEntity, tools)
    }
}
